#!/usr/bin/env python3
"""Potential data training using the MLP architecture.
Refactored version of `pystops_ml.training.rps_training_slim`
"""
from dataclasses import dataclass
from pathlib import Path

import torch
from pymytools.diagnostics import DataLoader
from pymytools.diagnostics import DataSaver
from pymytools.diagnostics import DataTracker
from pymytools.diagnostics import is_dir
from pymytools.logger import draw_rule
from pymytools.logger import logging
from pymytools.logger import markup
from pymytools.logger import timer

from pyrfp.model_tools import DEPMSE
from pyrfp.model_tools import DMEPMSE
from pyrfp.model_tools import set_loss_funcs
from pyrfp.model_tools import set_lr_schedular
from pyrfp.training_model import NLinearLayerModel
from pyrfp.training_runner import Runner
from pyrfp.training_tools import HDFTrainignDataHandler
from pyrfp.training_tools import Stages
from pyrfp.training_tools import training_summary
from pyrfp.types import TrainingConfig


@dataclass
class TrainingBasis:
    """Training basis class"""

    config: TrainingConfig
    """Training configuration data"""

    def __post_init__(self):
        # Print summary
        training_summary(self.config)

        # ------------------------------------------------------------------ #
        # Check training data directory.
        self.data_dir = Path(self.config["data_dir"]).as_posix()
        self.model_dir = Path(self.config["model_dir"]).as_posix()
        if not is_dir(self.data_dir, create=False):
            raise NotADirectoryError(
                f"Training: a given data directory ({self.data_dir}) not found."
            )

        # Check model save location
        # If not exists, create folder
        is_dir(self.model_dir, create=True, verbose=1)

        # ------------------------------------------------------------------ #
        # Device setup
        # Cuda status
        # Set device accordingly
        self.device = self.config["device"]
        self.dtype = self.config["dtype"]

        # Add data handler
        self.dh = HDFTrainignDataHandler(device=self.device, dtype=self.dtype)
        # Add data saver
        self.ds = DataSaver(self.model_dir)
        # Add data loader
        self.dl = DataLoader(self.dtype, self.device)

        # Check number of workers for the trainloader
        self.num_workers = self.config["num_workers"]

        # ------------------------------------------------------------------ #
        # Training preps
        # A number of batchs actually used in the training
        self.n_batchs = self.config["n_batch"]
        # Set a number of main epoch
        self.n_epochs = self.config["n_epochs"]
        # A number of data chunk size loaded at once
        self.n_data_chunk = self.config["data_chunk_size"]
        # Restart flag
        self.tr_restart = self.config["restart"]

        # ------------------------------------------------------------------ #
        # Architecture setups
        self.tr_lr = self.config["lr"]
        self.tr_lr_schedular = self.config["lr_schedular"]

        self.act_funcs = self.config["act_funcs"]
        self.layers = self.config["layers"]
        self.dropout = self.config["dropout"]
        self.batch_norm = self.config["batch_norm"]
        # Defining loss is bit changed!
        self.loss = self.config["loss"]

        # ------------------------------------------------------------------ #
        # Data tracker
        self.tracker_log_dir = Path(self.model_dir, Path("./run/"))
        self.tracker_overwrite = self.config["tracker_overwrite"]

        self.tracker = DataTracker(
            str(self.tracker_log_dir), overwrite=self.tracker_overwrite
        )

        # Timer
        self.timer = timer


class ParticleTraining(TrainingBasis):
    """Training only runs at the local system. Therefore, intended to work
    one CUDA device or local CPUs.
    """

    def __init__(self, config: TrainingConfig):
        super().__init__(config)

    def get_file_lists(
        self, is_validation: bool, verbose: int = 0
    ) -> tuple[list[str], list[str]]:
        """Get data file lists with a certain pattern.

        Args:
            is_validation (bool): flag to decide which file list returned.
            verbose (int): report level flag

        Returns:
            tuple[List[str], List[str]]: input and target file lists
        """

        if is_validation is False:
            data_dir = Path(self.data_dir, "training/").as_posix()
        else:
            data_dir = Path(self.data_dir, "validation/").as_posix()

        input_list, target_list = self.dh.get_file_list(data_dir)

        if verbose > 0:
            msg = markup(
                f"from {data_dir}, loading {len(input_list)} input files and {len(target_list)} target files",
                "blue",
                "bold",
            )
            logging.info(msg)

        return input_list, target_list

    def build_architecture(self, input_size: int, target_size: int):
        """Build MLP architecture based on `pyrfp.training_model.NLinearLayerModel`

        Args:
            input_size (int): input size
            target_size (int): target size

        Returns:
            torch.nn.Module: Neural net architecture
        """

        # construct SNN architecture

        logging.info(markup("configuring ann architecture...", "blue", "bold"))
        architecture = NLinearLayerModel(
            target_size,
            input_size,
            self.layers,
            self.act_funcs,
            batch_norm=self.batch_norm,
            dropout_rate=self.dropout,
            dtype=self.dtype,
        )

        if self.tr_restart is not None:
            # This only work with state dict. No entire model load

            assert isinstance(
                self.tr_restart, str
            ), "Training: restart option must be a model path"

            state_dict_loc = Path(self.model_dir, Path(self.tr_restart)).as_posix()

            logging.info(
                markup(
                    f"ATTENTION! restart training from {state_dict_loc}",
                    "yellow",
                    "bold",
                )
            )

            architecture = self.dl.read_state_dict(architecture, state_dict_loc)

        architecture.to(self.device)
        architecture.train()

        return architecture

    def set_schedular(
        self,
        optimizer: torch.optim.Optimizer,
    ) -> tuple[torch.optim.lr_scheduler._LRScheduler | None, float, int]:
        """Set lr schedular."""

        # Learning-rate schedular
        schedular = set_lr_schedular(
            self.tr_lr_schedular["name"],
            optimizer,
            self.tr_lr_schedular["options"],
        )

        if schedular is not None:
            # Enable schedular after schedular_start_after*n_epochs.
            # If option is not given, use 30% as default
            try:
                start_after = self.tr_lr_schedular["options"]["start_after"]
            except KeyError:
                start_after = 0.3

                msg = markup(
                    f"Schedular start_after value entry not given! Will use default value {start_after}!",
                    "magenta",
                    "bold",
                )
                logging.warning(msg)

            try:
                update_every = self.tr_lr_schedular["options"]["update_every"]
            except KeyError:
                update_every = 10

                msg = markup(
                    f"Schedular update_every value entry not given! Will use default value {update_every}!",
                    "magenta",
                    "bold",
                )
                logging.warning(msg)

        else:
            start_after = 1
            update_every = 1

        return schedular, start_after, update_every

    def excute(self) -> torch.nn.Module:
        """Excute training.

        Returns:
            torch.nn.Module: trained model
        """

        logging.info(markup("Preparing training..."), "blue", "bold")

        # Training data list
        tr_input_list, tr_target_list = self.get_file_lists(
            is_validation=False, verbose=1
        )
        # Validation data list
        val_input_list, val_target_list = self.get_file_lists(
            is_validation=True, verbose=1
        )

        # Get data size to construct architecture
        n_tr_files = len(tr_input_list)

        # Check one of data's size
        tr_input_size = self.dh.get_data_size(tr_input_list[0], "inputs")
        tr_target_size = self.dh.get_data_size(tr_target_list[0], "targets")

        tr_in_size = tr_input_size[1]
        tr_in_rows = tr_input_size[0]
        tr_out_size = tr_target_size[1]

        data_chunk = tr_in_rows if self.n_data_chunk is None else self.n_data_chunk

        # Validation data chunk size
        val_data_chunk = int(len(val_input_list) / n_tr_files * data_chunk)

        # Construct architecture
        architecture = self.build_architecture(tr_in_size, 9)

        # Optimizer
        optimizer = torch.optim.Adam(architecture.parameters(), lr=self.tr_lr)

        # Loss function
        criterion = set_loss_funcs(self.loss)

        (
            schedular,
            schedular_start_after,
            schedular_update_every,
        ) = self.set_schedular(optimizer)

        draw_rule(character="-")

        # Check nn architecture.
        logging.info(markup("checking MLP architecture...\n", "blue", "bold"))
        # Neural Net layer info
        print(architecture)
        # Optimizer info
        print(optimizer)
        # Scheduler info (if exists)
        if schedular is not None:
            schedular_name = self.tr_lr_schedular["name"]
            print(schedular_name + " (")
            for k, v in self.tr_lr_schedular["options"].items():
                print(f"    {k} : {v}")
            print(")")
        # Loss function info
        print(criterion)
        draw_rule(character="-")
        logging.info(markup("training starts! \n", "blue", "bold"))

        logging.info(markup(f"{str(self.device)} mode in use.", "blue", "bold"))

        self.timer.start("ann")

        saved = False

        try:
            for epoch in range(self.n_epochs):
                print("")

                tr_inputs, tr_targets = self.dh.get_input_target_from_dir(
                    tr_input_list, tr_target_list, data_chunk, False
                )
                val_inputs, val_targets = self.dh.get_input_target_from_dir(
                    val_input_list, val_target_list, val_data_chunk, True
                )

                # Training loader
                tr_tl = self.dh.get_trainlaoder(
                    tr_inputs, tr_targets, self.n_batchs, self.num_workers
                )

                # Validation loader
                val_tl = self.dh.get_trainlaoder(
                    val_inputs, val_targets, val_data_chunk, True
                )

                # Construct runners with loader
                training_runner = Runner(
                    tr_tl, architecture, criterion, optimizer, self.device
                )
                # Validation process doesn't require optimizer
                validation_runner = Runner(
                    val_tl, architecture, criterion, None, self.device
                )

                # Need better idea...
                if self.loss["name"] == "depmse" or self.loss["name"] == "dmepmse":
                    # Update factors (and skip epoch == 0)
                    if epoch % self.loss["options"]["update_every"] == 0 and epoch > 1:
                        assert isinstance(criterion, DEPMSE | DMEPMSE)
                        criterion.update_weight()

                    # Double check..
                    logging.info(
                        markup(
                            f"using loss_ep weight of {training_runner.compute_loss.weight}",
                            "yellow",
                            "bold",
                        )
                    )

                # Training starts!
                self.tracker.set_stage(Stages.TRAIN)
                training_runner.run(f"TRAIN |{epoch+1}/{self.n_epochs}|", self.tracker)

                self.tracker.set_stage(Stages.VAL)
                validation_runner.run(
                    f"VALID |{epoch+1}/{self.n_epochs}|", self.tracker
                )

                self.tracker.set_stage(Stages.NORMAL)
                self.tracker.add_epoch_metric(
                    "tr_count", training_runner.run_count, epoch
                )
                self.tracker.add_epoch_metric(
                    "val_count", validation_runner.run_count, epoch
                )

                # schedular resides outside since this is nested training
                # after 30% of n_epochs, decay learning rate
                if epoch > int(self.n_epochs * schedular_start_after):
                    # decay every 10 epochs
                    if epoch % schedular_update_every == 0 and schedular is not None:
                        print("")
                        schedular.step()

                # Save model during epoch
                # If self.config["save_every"] < 0, model will not be saved
                if (
                    epoch % self.config["save_every"] == 0
                    and epoch != 0
                    and self.config["save_every"] > 0
                ):
                    self.ds.save_model(architecture, f"model_{epoch}")

        except KeyboardInterrupt:
            saved = self.finalize(architecture)

        if saved is False:
            self.finalize(architecture)

        return architecture

    def finalize(self, architecture: torch.nn.Module) -> bool:
        """Finalize the training."""

        # Save model in case of keyboard interruption
        # In this case, model will be saved as "model.pth"
        self.ds.save_model(architecture, "model")

        self.timer.end("ann")
        elapsed_time = self.timer.elapsed("ann")

        logging.info(
            markup(rf"done in {elapsed_time} [s] :kissing_heart:", "green", "bold")
        )

        return True



#!/usr/bin/env python3
"""Particle simulator for the homogeneous relaxation.
It includes
    - DSMC
        - Most of implementation can be found in `pystops`. Here, only single species gridless DSMC is implemented to demonstrate our data-driven approach
        - I rewrote here (from `pystops`) since I want this package to be decoupled from `pystops`
    - FP
"""
import torch
from pymytools.constants import EPS0
from pymytools.constants import PI
from torch import Tensor


def binary_nanbu(
    vel_a: torch.Tensor,
    vel_b: torch.Tensor,
    m_a: float,
    m_b: float,
    q_a: float,
    q_b: float,
    n_b: float,
    ln_lambda: float,
    dt: float,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Collision according to elastic and binary collision."""

    dim = vel_a.size(0)

    # Relative velocity
    g = vel_a - vel_b

    # Magnitude of relative velocity and its perpendicular component
    g_mag = torch.sqrt(torch.sum(g**2, dim=1))
    g_perp = torch.sqrt(g[:, 1] * g[:, 1] + g[:, 2] * g[:, 2])

    mu = m_a * m_b / (m_a + m_b)

    s = (
        ln_lambda
        / (4 * PI)
        * ((q_a * q_b / (EPS0 * mu)) ** 2)
        * n_b
        / (g_mag**3)
        * dt
    )

    cos_xi = _nanbu_scattering(s)
    sin_xi = torch.sqrt(1 - cos_xi**2)

    h = torch.zeros_like(vel_a)
    rnd = 2 * PI * torch.rand(dim, dtype=vel_a.dtype)

    # Binary collision
    h[:, 0] = g_perp * torch.cos(rnd)
    h[:, 1] = (
        -(g[:, 1] * g[:, 0] * torch.cos(rnd) + g_mag * g[:, 2] * torch.sin(rnd))
        / g_perp
    )
    h[:, 2] = (
        -(g[:, 2] * g[:, 0] * torch.cos(rnd) - g_mag * g[:, 1] * torch.sin(rnd))
        / g_perp
    )

    vel_a -= mu / m_a * (g * (1 - cos_xi.view(dim, 1)) + h * sin_xi.view(dim, 1))
    vel_b += mu / m_a * (g * (1 - cos_xi.view(dim, 1)) + h * sin_xi.view(dim, 1))

    return (vel_a, vel_b)


def _nanbu_scattering(s_in: Tensor) -> Tensor:
    """Calculate scattering angle based on K.Nanbu's scheme (1997).

    Note:
        - Polynomial approximation is adopted from J. Derouillat et al. (2019)
    """

    cos_xi = torch.zeros_like(s_in)
    rand = torch.rand_like(cos_xi)

    # Make sure cos(xi) > 0
    rand[rand.le(0.001)] = 0.001

    mask_1 = s_in.lt(0.1)
    mask_2 = s_in.ge(0.1) & s_in.lt(3.0)
    mask_3 = s_in.ge(3.0) & s_in.lt(6.0)
    mask_4 = s_in.ge(6.0)

    # s < 0.1
    cos_xi[mask_1] = 1.0 + s_in[mask_1] * torch.log(rand[mask_1])

    # s >= 0.1 and s < 3.0
    inv_a = (
        0.00569578
        + (
            0.95602
            + (
                -0.508139
                + (0.479139 + (-0.12789 + 0.0238957 * s_in[mask_2]) * s_in[mask_2])
                * s_in[mask_2]
            )
            * s_in[mask_2]
        )
        * s_in[mask_2]
    )
    a = 1.0 / inv_a
    cos_xi[mask_2] = inv_a * torch.log(
        torch.exp(-a) + 2.0 * rand[mask_2] * torch.sinh(a)
    )

    # s >= 0.3 and s < 0.6
    a = 3.0 * torch.exp(-s_in[mask_3])
    inv_a = 1.0 / a
    cos_xi[mask_3] = inv_a * torch.log(
        torch.exp(-a) + 2.0 * rand[mask_3] * torch.sinh(a)
    )

    # s >= 0.6
    cos_xi[mask_4] = 2.0 * rand[mask_4] - 1.0

    return cos_xi



#!/usr/bin/env python3
"""Training runner"""
from typing import Any

import torch
from pymytools.diagnostics import DataTracker
from pymytools.progress import ProgressBar
from torch.utils.data.dataloader import DataLoader

from pyrfp.model_tools import ensemble_de_dt
from pyrfp.model_tools import ensemble_e
from pyrfp.model_tools import LOSS_CUSTOM
from pyrfp.training_tools import Metric


class Runner:
    def __init__(
        self,
        loader: DataLoader,
        model: torch.nn.Module,
        criterion: Any,
        optimizer: torch.optim.Optimizer | None,
        device: torch.device,
    ):
        """Run training.

        Args:
            loader (DataLoader): torch data loader
            model (torch.nn.Module): torch neural net architecture
            criterion (Any): loss function
            optimizer (Optional[torch.optim.Optimizer]): torch optimizer
                If it is None, runner runs for the validation
            device (torch.device): device to run the model
        """

        self.run_count = 0
        self.loader = loader
        self.loader_size = len(self.loader)
        self.model = model
        self.loss_function_name = criterion.__class__.__name__.lower()
        self.compute_loss = criterion
        self.optimizer = optimizer
        self.device = device
        self.accuracy_metric = Metric()

        self.target_batches: list[list[Any]] = []
        self.pred_batches: list[list[Any]] = []

        self.tol = 1e-1

    def run(self, desc: str, tracker: DataTracker) -> None:
        """Run training or validation.

        Args:
            desc (str): training description.
            tracker (Tracker): training tracker.
        """

        pbar = ProgressBar(self.loader, desc=desc, mininterval=0)

        for data, target in pbar:
            # forward
            loss, acc_abs, acc_rel, acc_eng = self._run_single(data, target)

            # Compute metrics
            tracker.add_batch_metric("loss", loss, self.run_count)
            tracker.add_batch_metric("accuracy_abs", acc_abs, self.run_count)
            tracker.add_batch_metric("accuracy_rel", acc_rel, self.run_count)

            if acc_eng is not None:
                tracker.add_batch_metric("accuracy_eng", acc_eng, self.run_count)
                desc_mod = desc + f" Acc-eng: {acc_eng:.2e} "
            else:
                desc_mod = desc + f" Acc-rel: {acc_rel:.2e} "

            if self.optimizer is not None:
                desc_mod += f" lr: {self.optimizer.param_groups[0]['lr']:.2e}"  # type: ignore
            else:
                desc_mod += f"             "
            pbar.set_description(desc_mod)

    def _run_single(
        self, data: torch.Tensor, target: torch.Tensor
    ) -> tuple[float, float, float, float | None]:
        """Feed forward process. This part is separated to handle duplication
        both in training and validation process.

        """
        self.run_count += 1

        if self.device == "cuda":
            data = data.to(self.device, non_blocking=True)
            target = target.to(self.device, non_blocking=True)
        else:
            data, target = data.to(self.device), target.to(self.device)

        batch_size = target.numel()

        # Inference
        prediction = self.model(data)

        target_size = target.shape[1]

        if self.loss_function_name in LOSS_CUSTOM:
            if target_size <= 9:
                msg = f"Training: target size has to be 12. But received {target_size}!"

                raise RuntimeError(msg)

            target_coeffs = target[:, :-3]
            wiener = target[:, -3::]

            # NOTE: Need to add dm/dt
            if self.loss_function_name == "depmse":
                loss = self.compute_loss(data, prediction, target_coeffs)
            else:
                loss = self.compute_loss(data, prediction, target_coeffs, wiener)

            if self.loss_function_name == "depmse":
                batch_accuracy_eng = ensemble_e(data, prediction).cpu().detach().numpy()
            else:
                batch_accuracy_eng = (
                    ensemble_de_dt(data, prediction, wiener).cpu().detach().numpy()
                )
        else:
            if target_size == 12:
                target_coeffs = target[:, :-3]
            elif target_size == 9:
                target_coeffs = target
            else:
                msg = (
                    "Wrong target size!. Should be either 9 or 12 "
                    f"but received {target_size}"
                )
                raise RuntimeError(msg)

            loss = self.compute_loss(prediction, target_coeffs)
            batch_accuracy_eng = None

        data_size = target_coeffs.numel()

        # percentage of diff below tolerance...
        diff = torch.abs(torch.div((target_coeffs - prediction), target_coeffs))
        batch_accuracy_abs = diff.sum().cpu().detach().numpy() / data_size

        batch_accuracy_rel = (diff < self.tol).sum().cpu().detach().numpy() / data_size

        self.accuracy_metric.update(batch_accuracy_rel, batch_size)

        # Back propagation
        if self.optimizer:
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

        batch_loss = loss.cpu().detach().numpy()

        return (
            batch_loss,
            batch_accuracy_abs,
            batch_accuracy_rel,
            batch_accuracy_eng,
        )



#!/usr/bin/env python3
"""Tools used in training"""
import random
from dataclasses import dataclass
from dataclasses import field
from enum import auto
from enum import Enum
from typing import Protocol

import torch
from pymytools.diagnostics import DataLoader
from pymytools.diagnostics import file_list_by_pattern
from pymytools.logger import draw_rule
from pymytools.logger import logging
from pymytools.logger import markup
from torch import Tensor
from torch.utils.data.dataloader import DataLoader as TorchDataLoader

from pyrfp.types import TrainingConfig


@dataclass
class Metric:
    """Contains metrics for the trained model."""

    values: list[float] = field(default_factory=list)
    running_total: float = 0.0
    num_updates: float = 0.0
    average: float = 0.0

    def update(self, value: float, batch_size: int):
        """Update metrics."""

        self.values.append(value)
        self.running_total += value * batch_size
        self.num_updates += batch_size
        self.average = self.running_total / self.num_updates


class Stages(Enum):
    """Representation of training stages. Either TRAIN or VAL (validation)."""

    TRAIN = auto()
    VAL = auto()
    NORMAL = auto()


class TrainingTracker(Protocol):
    """Protocol for the Tensorboard tracker `pymytools.diagnostics.DataTracker`."""

    def add_epoch_metric(self, name: str, value: float, step: int) -> None:
        """Implements logging a epoch-level metric."""

    def add_batch_metric(self, name: str, value: float, step: int) -> None:
        """Add batch metric to tensorboard."""

    def add_hparams(
        self, hparams: dict[str, str | float], metrics: dict[str, float]
    ) -> None:
        """Implements logging hyperparamters."""

    def set_stage(self, stage: Stages) -> None:
        """Sets the stage of the tracker."""

    def flush(self) -> None:
        """Flushes the stage of the tracker."""


def training_summary(config: TrainingConfig) -> None:
    """Print out training summary"""

    logging.info(markup("Summary", "red", "bold"))
    draw_rule(character="-")
    for k, v in config.items():
        logging.info(markup(f"{k}: {v}", "blue", "bold"))
    draw_rule(character="-")


@dataclass
class HDFTrainignDataHandler:
    device: torch.device = torch.device("cpu")
    """Device to use for data loading."""
    dtype: torch.dtype = torch.float32

    def __post_init__(self) -> None:
        self.dl = DataLoader(device=self.device, dtype=self.dtype)
        """Data loader. Dtype of the returned data is `torch.float64` and device is `self.device`."""

    @property
    def is_cuda(self) -> bool:
        """Cuda identifier"""
        return self.device.type == "cuda"

    def get_data_size(self, data_dir: str, addr: str) -> torch.Size:
        """Return data size (`idx`'s column length) of the training data.

        Args:
            data_dir (str): Directory where the training data is stored.
            addr (str): Address used in HDF5 format. Should be posix style.
            idx (int): Column index of the data to be loaded.
        """

        # If address comes with `/`, split it and take the last part.
        dataset_name = addr.rsplit("/", 1)[-1]
        loaded_data = self.dl.read_hdf5(data_dir, dataset_name)[dataset_name]

        return loaded_data.shape

    def get_data_by_chunk(
        self, file_path: str, addr: str, chunk: torch.Tensor
    ) -> Tensor:
        """Load HDF5 data by chunk (`torch.Tensor`)."""

        return self.dl.read_hdf5(file_path, addr)[addr][chunk, :]

    def get_file_list(self, data_dir: str) -> tuple[list[str], list[str]]:
        """Return input and target file list (by pattern)."""

        return file_list_by_pattern(data_dir, "input_*.h5"), file_list_by_pattern(
            data_dir, "target_*.h5"
        )

    def get_input_target_from_dir(
        self,
        input_list: list[str],
        target_list: list[str],
        chunk_size: int,
        is_validation: bool = False,
        ratio: float = 1.0,
    ) -> tuple[Tensor, Tensor]:
        """From the given input and target file list, load each data by a chunk size/len(input_list) and construct torch trainingloader.

        Args:
            input_list (list[str]): list contains input data pathes
            target_list (list[str]): list contains target data pathes
            chunk_size (int): data chunk size to be returned. If negative, load all data in a file.
            is_validation (bool, optional): If true, this is validation data. Defaults to False.
            ratio (float, optional): Ratio of data to be loaded. Used to adjust total number of files to be loaded. Defaults to 1.0.

        Returns:
            tuple[Tensor, Tensor]: loaded tensor of input and target
        """

        assert len(input_list) == len(
            target_list
        ), "DataHandler: input and target list should have the same length."

        assert ratio <= 1.0, "DataHandler: ratio should be less than or equal to 1.0."

        n_files = int(len(input_list) * ratio)

        if n_files == 0:
            n_files = 1

        chunk_size = int(chunk_size / n_files)

        # Shuffle input and target list to avoid bias
        random.shuffle(input_list)
        random.shuffle(target_list)

        if is_validation:
            prefix = "LOAD DATA - VALID: "
        else:
            prefix = "LOAD DATA - TRAIN: "

        msg = markup(
            prefix + f"from {n_files} files, loading chunk size of ({chunk_size}) each",
            "blue",
            "bold",
        )
        logging.info(msg)

        inputs: list[Tensor] = []
        targets: list[Tensor] = []

        for i, t in zip(input_list[:n_files], target_list[:n_files]):
            column_size = self.get_data_size(i, "inputs")

            assert (
                chunk_size <= column_size[0]
            ), f"DataHandler: chunk size ({chunk_size}) should be less than or equal to the column size of the given data ({column_size[0]})."

            if chunk_size > 0:
                # chunk = torch.randperm(column_size[0], device=self.device)[:chunk_size]
                chunk = torch.randperm(column_size[0], device=torch.device("cpu"))[
                    :chunk_size
                ]
                chunk.to(self.device)

            else:
                # chunk = torch.randperm(column_size[0], device=self.device)
                chunk = torch.randperm(column_size[0], device=torch.device("cpu"))
                chunk.to(self.device)

            inputs.append(self.get_data_by_chunk(i, "inputs", chunk))
            targets.append(self.get_data_by_chunk(t, "targets", chunk))

        return torch.vstack(inputs), torch.vstack(targets)

    def get_trainlaoder(
        self, inputs: Tensor, targets: Tensor, batch_size: int, num_workers: int
    ) -> TorchDataLoader:
        """From inputs and targets, construct torch Dataloader.

        Args:
            inputs (Tensor): Input data
            targets (Tensor): Target data
            batch_size (int): Batch size
            num_workers (int): Number of workers. Currently, only used when device is cuda. (in cpu mode, utilize only 1 processor)
        """

        td = torch.utils.data.TensorDataset(inputs, targets)  # type: ignore

        # Set pin memory to True if device is cuda
        pm = True if self.is_cuda else False
        dl = TorchDataLoader(
            td,
            batch_size=batch_size,
            pin_memory=pm,
            shuffle=True,
            num_workers=num_workers if self.is_cuda else 0,
        )

        return dl



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyrfp']

package_data = \
{'': ['*']}

install_requires = \
['pyapes>=0.2.1,<0.3.0', 'pymaxed>=0.1.7,<0.2.0', 'pymytools>=0.1.12,<0.2.0']

setup_kwargs = {
    'name': 'pyrfp',
    'version': '0.1.2',
    'description': 'Python package for the data-driven Rosenbluth-Fokker-Planck equation.',
    'long_description': "# Python Package for the Rosenbluth Fokker Planck (RFP) Equation\n\n> Currently, heavily under renovation (refactoring) from my old code (in my other private repository)\n\n## Description\n\nThis package is refactored version of a part of the `pystops_ml` code.\nI've separated only data generation and training part from `pystops_ml`.\n\nUnlike `pystops_ml`, this module doesn't utilize distributed training. (`DDP` feature not needed)\n\nThis code is part of my paper, Data-Driven Stochastic Particle Scheme for Collisional Plasma Simulations.\n\n- Preprint is available at [SSRN](https://ssrn.com/abstract=4108990)\n\n## Features\n\n- Data generation\n  - Uncertainty quantification using the maximum entropy distribution (using `pymaxed`)\n  - Axisymmetric evaluation of the Rosenbluth potentials and their derivatives\n- Data training: supports `cpu`, `cuda`, and `mps` training\n- Data-driven simulation: `rfp-ann`\n\n## Installation\n\nYou can install via `pip`\n\n```bash\n$python3 -m pip install pyrfp\n```\n\n## Dependencies\n\n- Global\n- `python >=3.10`\n- `torch >= 1.13.1`\n\n- Personal project\n  - `pymaxed` (for the Maximum Entropy Distribution)\n  - `pyapes` (for the field solver)\n  - `pymytools` (miscellaneous tools including data IO, logging, and progress bar)\n",
    'author': 'Kyoungseoun Chung',
    'author_email': 'kyoungseoun.chung@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<3.12',
}


setup(**setup_kwargs)
